/*
 * timingSqrtHeron_g.cu - Bailey style testing for square root
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2013 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu valentina.popescu@univ-tlse3.fr
 *               Mioara Joldes joldes@laas.fr
 */
/* Test file for square root of mprec using doubles */

#include <ctime>
#include <cstdio>
#include <gmp.h>
#include <iostream>
#include <iomanip>
#include <string>
using namespace std;
#include <gmp.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <mpfr.h>
#include <mpfi.h>
#include "mpfi_io.h"
#include "../src/multi_prec.h"
#include <cuda_runtime.h>
#include <helper_functions.h>
#include <helper_cuda.h>
#define prec 4 //prec 
#define precu 4 //prec result

/*We test the performance of multi_prec reciprocals by taking random generated nonoverlapping expansions and then adding them;
The point is to get performance measurements in MFlops for the following combinations:
-- 1d/2d
-- 2d/2d
-- 4d/2d
-- etc;

-- 1d/4d
-- 2d/4d
-- 4d/4d
-- etc.
*/

const double MICRO(1e-6);       // Used for the timing.
//----------------------------------------------------------------------
// A basic timer that works for OMP code.
//----------------------------------------------------------------------
double my_timer(timeval &tim, double &t_start, double &t_stop,
                          const string &command){
  if ( 0 == command.compare("start") ) {
  	gettimeofday(&tim, NULL);
    t_start = tim.tv_sec+(tim.tv_usec*MICRO);
  }
  if ( 0 == command.compare("stop") ) {
    gettimeofday(&tim, NULL);
    t_stop = tim.tv_sec+(tim.tv_usec*MICRO);
    cout << "Wall time: " << setprecision(8) << fixed
         << t_stop - t_start << " seconds.";
  }
  return (t_stop - t_start);
}

//--------------------------------------------------------------------------
// Return an array of CUDA deviceIDs that are available and have compute 
// capability at least 2.0.
int* getCudaDevices20(int &NUM_GPU){
  cudaDeviceProp deviceProp;
  deviceProp.major = 0;
  deviceProp.minor = 0;
  int n, version;
  cudaError_t cudaStatus;
  checkCudaErrors( cudaGetDeviceCount(&n) );
  int* deviceArray;
  deviceArray=(int*)malloc(n*sizeof(int));
  int deviceID;
  int k=0;
  for (deviceID=0;deviceID<n;deviceID++){
    cudaStatus=cudaSetDevice(deviceID);
    if (cudaStatus ==cudaSuccess) {
      checkCudaErrors( cudaGetDeviceProperties(&deviceProp, deviceID) );
      version = deviceProp.major * 0x10 + deviceProp.minor;
      if (version >= 0x20) {
      	deviceArray[k]=deviceID;
        k++;
      }
    }
  }
  if ( k == 0 ){
  	printf("There is no device supporting a minimum CUDA compute capability 2.0\n");
  } else {
    printf("Found %d/%d devices supporting a minimum CUDA compute capability 2.0\n", k,n);
    int dev;
    cudaSetDevice( dev = gpuGetMaxGflopsDeviceId() );
    checkCudaErrors( cudaChooseDevice(&dev, &deviceProp) );
    checkCudaErrors( cudaGetDeviceProperties(&deviceProp, dev) );
    printf("CUDA device [%s] has %d Multi-Processors, Compute %d.%d\n\n",
           deviceProp.name, deviceProp.multiProcessorCount,
           deviceProp.major, deviceProp.minor);
    //printf("SharedMemPerBlock %ld bytes, TotalGlobalMem %ld bytes\n\n", 
    //deviceProp.sharedMemPerBlock, deviceProp.totalGlobalMem);    
  }
  cudaDeviceReset();
  NUM_GPU = k;
  return deviceArray;
}

int* printandGetInfo(int &NUM_GPU, int BLOCKS, int THREADS){
//	cout << endl << "**********************************************************" << endl
//       << endl << "Latest compilation: " << __DATE__ << ", " << __TIME__ << endl;

	int* deviceArray = getCudaDevices20(NUM_GPU);

  cout << "Computing with " << BLOCKS << " blocks, each having " << THREADS << " threads." << endl;
  return deviceArray;
}


__global__ void testMulGPU(const double* a, double* b, int trials){
	multi_prec<prec,double> ma;
	multi_prec<precu,double> mb;
	ma.setData(a, prec);
	mb = 1.0;

	for (int i=0; i<trials; i++){
    naiveSqrtHeron(mb,ma);
  }
	for (int i=0; i<precu; i++)
		b[i] = mb[i];

  __syncthreads();
}


int main(int argc, char *argv[]) {
	if(argc < 2){
		cout << "Usage: ./timingMul_g trials" << endl; 
   	return 0;
 	}
	int trials = atoi(argv[1]);
	char hostname[1024];
  gethostname(hostname, 1024);
  cout << "Latest compilation: " << __DATE__ << ", " << __TIME__ << " on " << hostname << endl;

	int THREADS = 1;
	int BLOCKS = 1;
  int NUM_GPU(1);

	//get the deviceIndex of available GPUs with capability>=2.0
  int *deviceArray = printandGetInfo(NUM_GPU, BLOCKS, THREADS);

	double *ad = new double[prec];
	double *resd = new double[precu];

	multi_prec<prec, double> mpa;
  double buff[prec];
  generateNormalizedExpansion<prec>(buff);
  mpa.setData(buff,prec);
  mpa = 1.0 + (1.0/trials) * mpa;
	for(int i=0; i<prec; i++)
		ad[i] = mpa[i]; 

	cudaError_t cudaStatus;
	//get the first device compatible
	cudaStatus = cudaSetDevice(deviceArray[0]);
	if(cudaStatus == cudaErrorInvalidDevice){
  	printf("CUDA ERROR: %s\n", cudaGetErrorString(cudaStatus) );
    return (int)cudaStatus;
  }

	double *dev_a, *dev_b;
  checkCudaErrors(cudaMalloc((void**)&dev_a, prec * sizeof(double)));
	checkCudaErrors(cudaMalloc((void**)&dev_b, precu * sizeof(double)));

	checkCudaErrors(cudaMemcpy(dev_a, ad, prec * sizeof(double),cudaMemcpyHostToDevice));

	cudaEvent_t start, stop;
  checkCudaErrors(cudaEventCreate(&start));
  checkCudaErrors(cudaEventCreate(&stop));
  checkCudaErrors(cudaEventRecord(start,0));

	testMulGPU<<<BLOCKS, THREADS>>>(dev_a, dev_b, trials);
	cudaDeviceSynchronize();
	cudaError_t error = cudaGetLastError();
  if(error!=cudaSuccess){
  	printf("CUDA KERNEL ERROR: %s\n", cudaGetErrorString(error));
  }else{
		checkCudaErrors(cudaEventRecord(stop,0));
    checkCudaErrors(cudaEventSynchronize(stop));
    float elapsedTime;
		checkCudaErrors(cudaEventElapsedTime(&elapsedTime,start,stop));
		
		cout << "Testing " << trials << " square roots with " << precu << "d = sqrt(" << prec << "d) precision." << endl;
    printf("CUDA time: %3.1f ms\n", elapsedTime);
    cout << "CUDA Performance: " << (1e-6*trials*1000)/elapsedTime << " Mops/second." << endl;
		checkCudaErrors(cudaEventDestroy(start));
    checkCudaErrors(cudaEventDestroy(stop));
		checkCudaErrors(cudaMemcpy(resd, dev_b, precu*sizeof(double),cudaMemcpyDeviceToHost));
		double x = 0.0;
		for (int i=0; i<precu; i++){
			x = x + resd[i];
		}
		cout<< "Result: " << x << endl;
	}

	cudaFree(dev_a);
	cudaFree(dev_b);

	return 0;
}

