/*
 * FP_basicOp.h
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2014 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */
/*
 * Specialization for basic FP operations
 */

#ifndef _FPbasicOp_h
#define _FPBasicOp_h

/*****************************************************************************************/
/** Templated function for adding two FP numbers, even if they don't have the same type **/
/*****************************************************************************************/
template<class T1, class T2, class TR>
__device__ static __forceinline__ TR FPadd_rn_gpu(const T1 x, const T2 y){
	return x + y;
}

template<>
__device__ __forceinline__ double FPadd_rn_gpu<double,double,double>(const double x, const double y){
  return __dadd_rn(x, y);
}

template<>
__device__ __forceinline__ double FPadd_rn_gpu<double,float,double>(const double x, const float y){
  return __dadd_rn(x, (double)y);
}

template<>
__device__ __forceinline__ double FPadd_rn_gpu<float,double,double>(const float x, const double y){
  return __dadd_rn((double)x, y);
}

template<>
__device__ __forceinline__ double FPadd_rn_gpu<float,float,double>(const float x, const float y){
  return __dadd_rn((double)x, (double)y);
}

template<>
__device__ __forceinline__ float FPadd_rn_gpu<float,float,float>(const float x, const float y){
  return __fadd_rn(x, y);
}

template<>
__device__ __forceinline__ float FPadd_rn_gpu<float,double,float>(const float x, const double y){
  return __fadd_rn(x, (float)y);
}

template<>
__device__ __forceinline__ float FPadd_rn_gpu<double,float,float>(const double x, const float y){
  return __fadd_rn((float)x, y);
}

template<>
__device__ __forceinline__ float FPadd_rn_gpu<double,double,float>(const double x, const double y){
  return __fadd_rn((float)x, (float)y);
}

template<class T1, class T2, class TR>
__host__ static __forceinline__ TR FPadd_rn_cpu(const T1 x, const T2 y){
	return x + y;
}

template<class T1, class T2, class TR>
__host__ __device__ static __forceinline__ TR FPadd_rn(const T1 x, const T2 y){
  #if defined __CUDA_ARCH__
    return FPadd_rn_gpu<T1,T2,TR>(x, y);
  #else
    return FPadd_rn_cpu<T1,T2,TR>(x, y);
  #endif
}

/**********************************************************************************************/
/** Templated function for multiplying two FP numbers, even if they don't have the same type **/
/**********************************************************************************************/
template<class T1, class T2, class TR>
__device__ static __forceinline__ TR FPmul_rn_gpu(const T1 x, const T2 y){
	return x * y;
}

template<>
__device__ __forceinline__ double FPmul_rn_gpu<double,double,double>(const double x, const double y){
  return __dmul_rn(x, y);
}

template<>
__device__ __forceinline__ double FPmul_rn_gpu<double,float,double>(const double x, const float y){
  return __dmul_rn(x, (double)y);
}

template<>
__device__ __forceinline__ double FPmul_rn_gpu<float,double,double>(const float x, const double y){
  return __dmul_rn((double)x, y);
}

template<>
__device__ __forceinline__ double FPmul_rn_gpu<float,float,double>(const float x, const float y){
  return __dmul_rn((double)x, (double)y);
}

template<>
__device__ __forceinline__ float FPmul_rn_gpu<float,float,float>(const float x, const float y){
  return __fmul_rn(x, y);
}

template<>
__device__ __forceinline__ float FPmul_rn_gpu<float,double,float>(const float x, const double y){
  return __fmul_rn(x, (float)y);
}

template<>
__device__ __forceinline__ float FPmul_rn_gpu<double,float,float>(const double x, const float y){
  return __fmul_rn((float)x, y);
}

template<>
__device__ __forceinline__ float FPmul_rn_gpu<double,double,float>(const double x, const double y){
  return __fmul_rn((float)x, (float)y);
}

template<class T1, class T2, class TR>
__host__ static __forceinline__ TR FPmul_rn_cpu(const T1 x, const T2 y){
  return x * y;
}

template<class T1, class T2, class TR>
__host__ __device__ static __forceinline__ TR FPmul_rn(const T1 x, const T2 y){
  #if defined __CUDA_ARCH__
    return FPmul_rn_gpu<T1,T2,TR>(x, y);
  #else
    return FPmul_rn_cpu<T1,T2,TR>(x, y);
  #endif
}

/*************************************************************************************************/
/** Templated function for the FMA instruction, even if the FP numbers don't have the same type **/
/*************************************************************************************************/
template<class T1, class T2, class T3, class TR>
__device__ static __forceinline__ TR FPfma_rn_gpu(const T1 x, const T2 y, const T3 z){
	return fma(x, y, z);
}

template<>
__device__ __forceinline__ double FPfma_rn_gpu<double,double,double,double>(const double x, const double y, const double z){
  return __fma_rn(x, y, z);
}

template<>
__device__ __forceinline__ double FPfma_rn_gpu<float,float,double,double>(const float x, const float y, const double z){
  return __fma_rn((double)x, (double)y, z);
}

template<>
__device__ __forceinline__ double FPfma_rn_gpu<float,double,float,double>(const float x, const double y, const float z){
  return __fma_rn((double)x, y, (double)z);
}

template<>
__device__ __forceinline__ double FPfma_rn_gpu<double,float,float,double>(const double x, const float y, const float z){
  return __fma_rn(x, (double)y, (double)z);
}

template<>
__device__ __forceinline__ double FPfma_rn_gpu<float,double,double,double>(const float x, const double y, const double z){
  return __fma_rn((double)x, y, z);
}

template<>
__device__ __forceinline__ double FPfma_rn_gpu<double,float,double,double>(const double x, const float y, const double z){
  return __fma_rn(x, (double)y, z);
}

template<>
__device__ __forceinline__ double FPfma_rn_gpu<double,double,float,double>(const double x, const double y, const float z){
  return __fma_rn(x, y, (double)z);
}

template<>
__device__ __forceinline__ double FPfma_rn_gpu<float,float,float,double>(const float x, const float y, const float z){
  return __fma_rn((double)x, (double)y, (double)z);
}

template<>
__device__ __forceinline__ float FPfma_rn_gpu<double,double,double,float>(const double x, const double y, const double z){
  return __fmaf_rn((float)x, (float)y, (float)z);
}

template<>
__device__ __forceinline__ float FPfma_rn_gpu<float,float,double,float>(const float x, const float y, const double z){
  return __fmaf_rn(x, y, (float)z);
}

template<>
__device__ __forceinline__ float FPfma_rn_gpu<float,double,float,float>(const float x, const double y, const float z){
  return __fmaf_rn(x, (float)y, z);
}

template<>
__device__ __forceinline__ float FPfma_rn_gpu<double,float,float,float>(const double x, const float y, const float z){
  return __fmaf_rn((float)x, y, z);
}

template<>
__device__ __forceinline__ float FPfma_rn_gpu<float,double,double,float>(const float x, const double y, const double z){
  return __fmaf_rn(x, (float)y, (float)z);
}

template<>
__device__ __forceinline__ float FPfma_rn_gpu<double,float,double,float>(const double x, const float y, const double z){
  return __fmaf_rn((float)x, y, (float)z);
}

template<>
__device__ __forceinline__ float FPfma_rn_gpu<double,double,float,float>(const double x, const double y, const float z){
  return __fmaf_rn((float)x, (float)y, z);
}

template<>
__device__ __forceinline__ float FPfma_rn_gpu<float,float,float,float>(const float x, const float y, const float z){
  return __fmaf_rn(x, y, z);
}

/*@ requires xy == \round_double(\NearestEven,x*y) && 
  @          \abs(x) <= 0x1.p995 && 
  @          \abs(y) <= 0x1.p995 &&
  @          \abs(x*y) <=  0x1.p1021;
  @ ensures  ((x*y == 0 || 0x1.p-969 <= \abs(x*y)) 
  @                 ==> x*y == xy+\result);
  @*/
/* Dekker algorithm */
__host__ static __forceinline__ double FPfma_rn_cpu(const double x, const double y, double xy) {
  double C,px,qx,hx,py,qy,hy,lx,ly,fma;
//  C = 0x8000001p0;
  /*@ assert C == 0x1p27+1; */
	C = 0x1p27+1;

  px = x*C;
  qx = x-px;
  hx = px+qx;
  lx = x-hx;

  py = y*C;
  qy = y-py;
  hy = py+qy;
  ly = y-hy;

  fma = -x*y+hx*hy;
  fma += hx*ly;
  fma += hy*lx;
  fma += lx*ly;
  return fma;
}

__host__ static __forceinline__ float FPfma_rn_cpu(const float x, const float y, float xy) {
  float C,px,qx,hx,py,qy,hy,lx,ly,fma;
  C = 0x1p12 + 1;
  /*@ assert C == 0x1p27+1; */

  px = x*C;
  qx = x-px;
  hx = px+qx;
  lx = x-hx;

  py = y*C;
  qy = y-py;
  hy = py+qy;
  ly = y-hy;

  fma = -x*y+hx*hy;
  fma += hx*ly;
  fma += hy*lx;
  fma += lx*ly;
  return fma;
}

template<class T1, class T2, class T3, class TR>
__host__ __device__ static __forceinline__ TR FPfma_rn(const T1 x, const T2 y, const T3 z){
  #if defined __CUDA_ARCH__
    return FPfma_rn_gpu<T1,T2,T3,TR>(x, y, z);
  #else
		#ifdef FP_FAST_FMA
    #warning cpu has fma
			return fma(x, y, z);
		#else
    	return FPfma_rn_cpu(x, y, z);
		#endif
  #endif
}

#endif
