/*
 * addition.h
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2014 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */
/*
 * addition functions and their specialized definitions
 */

#ifndef _addition_h
#define _addition_h
#include "renorm.h"

/** Merges the arrays x of size K and y of size L, and puts the result in the array z
    All 3 arrays x, y and z are sorted in decreasing order of magnitude
		K - size of x, L - size of y **/
template <int K, class TK, int L, class TL, class TR>
__host__ __device__ static __forceinline__ void merge(TK const *x, TL const *y, TR *z){
  int i=0, j=0;
  for(int n=0; n<K+L; n++){
    if(i==K || (j<L && abs(y[j])>abs(x[i]))){ z[n] = y[j]; j++; }
		else{ z[n] = x[i]; i++; }
  }
}

/** Adds x and y and returns the result in r. Takes into account all the terms in x and y.
		Error bound: |x+y-r| <= 2ulp(r[R-1])
		K - size of x, L - size of y, R - size of r **/
template <int K, class TK, int L, class TL, int R, class TR>
__host__ __device__ static __forceinline__ void certifiedAdd(const TK *x, const TL *y, TR *r){
  TR f[K+L];
  merge<K,TK,L,TL,TR>(x,y,f);
  renorm2L<K+L,R,TR>(f);
  for(int i=0;i<R;i++) r[i]=f[i];
}

/** Adds x and y and returns the result in r. Takes into account only the most significant R terms in x and y.
    K - size of x, L - size of y, R - size of r **/
template <int K, class TK, int L, class TL, int R, class TR>
__host__ __device__ static __forceinline__ void truncatedAdd(const TK *x, const TL *y, TR *r){
  TR f[K+L];
  merge<(R<K)?R:K,TK,(R<L)?R:L,TL,TR>(x,y,f);
  renorm2L<((R<K)?R:K)+((R<L)?R:L),R,TR>(f);
  for(int i=0;i<R;i++) r[i]=f[i];
}

//*********************************************************//
/** Merge algorithm used by the certifiedMul function. The array sizes are given as parameters **/
template <class TK, class TL, class TR>
__host__ __device__ static __forceinline__ void merge_4Mul(TK const *x, const int K, TL const *y, const int L, TR *z){
  int i=0, j=0;
  for(int n=0; n<K+L; n++){
    if(i==K || (j<L && abs(y[j])>abs(x[i]))){ z[n] = y[j]; j++; }
    else{ z[n] = x[i]; i++; }
  }
}
/** Addition algorithm used by the certifiedMul function. The array sizes are given as parameters **/
template <class TK, class TL, class TR>
__host__ __device__ static __forceinline__ int certifiedAdd_4Mul(const TK *x, const int K, const TL *y, const int L, TR *r){
  TR *f = new TR[K+L];
  merge_4Mul<TK,TL,TR>(x,K,y,L,f);
  return renorm2L_4Mul<TR>(f,K+L,r);
}

//*********************************************************//
/** Addition algorithm used by the functions based on the Newton-Raphson iteration. The renormalization function does all the subsequent levels **/
template <int K, class TK, int L, class TL, int R, class TR>
__host__ __device__ static __forceinline__ void truncatedAdd_4Ntn(const TK *x, const TL *y, TR *r){
  TR f[K+L];
  merge<(R<K)?R:K,TK,(R<L)?R:L,TL,TR>(x,y,f);
  renorm3L<((R<K)?R:K)+((R<L)?R:L),R,TR>(f);
  for(int i=0; i<R; i++) r[i] = f[i];
}

//*********************************************************//
//************ specialized functions for merge ************//
//*********************************************************//
template <>
__host__ __device__ __forceinline__ void merge<1,double,1,double,double>(double const *x, double const *y, double *z){
	if(abs(y[0])>abs(x[0])){ z[0] = y[0]; z[1] = x[0];}
	else{ z[0] = x[0]; z[1] = y[0];}
}

//*********************************************************//
//******** specialized functions for certifiedAdd *********//
//*********************************************************//
template <>
__host__ __device__ __forceinline__ void certifiedAdd<1,double,1,double,1,double>(const double *x, const double *y, double *r){ r[0] = FPadd_rn<double,double,double>(x[0], y[0]); }
template <>
__host__ __device__ __forceinline__ void certifiedAdd<1,double,1,double,2,double>(const double *x, const double *y, double *r){ r[0] = two_sum<double,double,double>(x[0], y[0], r[1]); }

//*********************************************************//
//******** specialized functions for truncatedAdd *********//
//*********************************************************//
template <>
__host__ __device__ __forceinline__ void truncatedAdd<1,double,1,double,1,double>(const double *x, const double *y, double *r){ r[0] = FPadd_rn<double,double,double>(x[0], y[0]); }
template <>
__host__ __device__ __forceinline__ void truncatedAdd<1,double,1,double,2,double>(const double *x, const double *y, double *r){ r[0] = two_sum<double,double,double>(x[0], y[0], r[1]); }

//*********************************************************//
//****** specialized functions for truncatedAdd_4Ntn ******//
//*********************************************************//
template <>
__host__ __device__ __forceinline__ void truncatedAdd_4Ntn<1,double,1,double,1,double>(const double *x, const double *y, double *r){ r[0] = FPadd_rn<double,double,double>(x[0], y[0]); }
template <>
__host__ __device__ __forceinline__ void truncatedAdd_4Ntn<1,double,1,double,2,double>(const double *x, const double *y, double *r){ r[0] = two_sum<double,double,double>(x[0], y[0], r[1]); }

#endif
