/*
 * errorFreeTransf.h
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2014 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */
/*
 * Error Free Transforms implementation
 * 2Sum, 2Prod and variants
 */

#ifndef _errorFreeTransf_h
#define _errorFreeTransf_h
//#define WITH3SUM
/**** CPU_FMA: FP_FAST_FMA --constant in math.h,
      if defined, it indicates that fma(S) generally executes about as fast as,
      or faster than a multiply and an add of double operands.*****/ 

#include "FP_basicOp.h"

/***************************************************************/
/******************** ERROR FREE TRANSFORMS ********************/
/***************************************************************/
template<class T1, class T2, class TR>
__host__ __device__ static __forceinline__ TR quick_two_sum(const T1 a, const T2 b, TR &err){
  TR s = FPadd_rn<T1,T2,TR>(a, b);
  TR z = FPadd_rn<TR,T1,TR>(s, -a);
  err = FPadd_rn<T2,TR,TR>(b, -z);
  return s;
}

/* Computes fl(a-b) and err(a-b).  Assumes |a| >= |b| */
template<class T1, class T2, class TR>
__host__ __device__ static __forceinline__ TR quick_two_diff(const T1 a, const T2 b, TR &err){
  TR s = FPadd_rn<T1,T2,TR>(a, -b);
  TR z = FPadd_rn<T1,TR,TR>(a, -s);
  err = FPadd_rn<TR,T2,TR>(z, -b);
  return s;
}

/* Computes fl(a+b) and err(a+b).  */
template<class T1, class T2, class TR>
__host__ __device__ static __forceinline__ TR two_sum(const T1 a, const T2 b, TR &err){
	TR s = FPadd_rn<T1,T2,TR>(a, b);
  TR aa = FPadd_rn<TR,T2,TR>(s, -b);
  TR bb = FPadd_rn<TR,TR,TR>(s, -aa);
  TR da = FPadd_rn<T1,TR,TR>(a, -aa);
  TR db = FPadd_rn<T2,TR,TR>(b, -bb);
  err = FPadd_rn<TR,TR,TR>(da, db);
  return s;
}

/* Computes fl(a-b) and err(a-b).  */
template<class T1, class T2, class TR>
__host__ __device__ static __forceinline__ TR two_diff(const T1 a, const T2 b, TR &err){
  TR s = FPadd_rn<T1,T2,TR>(a, -b);
  TR bb = FPadd_rn<TR,T1,TR>(s, -a);
  TR aa = FPadd_rn<TR,TR,TR>(s, -bb);
  TR da = FPadd_rn<T1,TR,TR>(a, -aa);
  TR db = FPadd_rn<T2,TR,TR>(b, bb);
  err = FPadd_rn<TR,TR,TR>(da, -db);
  return s;
}

/* Computes fl(a*b) and err(a*b). */
template<class T1, class T2, class TR>
__host__ __device__ static __forceinline__ TR two_prod(const T1 a, const T2 b, TR &err){
	const TR p = FPmul_rn<T1,T2,TR>(a, b);
 	err = FPfma_rn<T1,T2,TR,TR>(a, b, -p);

 	return p;
}

/*__device__ static __forceinline__ double round2odd_gpu(const double a, const double b){
  double d = __dadd_rd(a, b);
  double u = __dadd_ru(a, b);
  double e = __dadd_rn(u, d);
  e = __dmul_rn(e, 0.5);
  return __dadd_rn(__dsub_rn(u, e), d);
}

__device__ static __forceinline__ float round2odd_gpu(const float a, const float b){
  float d = __fadd_rd(a, b);
	float u = __fadd_ru(a, b);
  float e = __fadd_rn(u, d);
  e = __fmul_rn(e, 0.5);
  return __fadd_rn(__fsub_rn(u, e), d);
}

template<class T>
__host__ static inline T round2odd_cpu(const T a, const T b){
  T* thp;
  T th, tl;
  thp = (T*)malloc(sizeof(T));	
  th = quick_two_sum_cpu(a, b, tl);
  *thp = th;
  long long *thpb = (long long*)thp;
  //round to odd if tl is not zero
  if(tl!=0.0){
    //if the mantissa is odd, there is nothing to be done.
		if (!(*thpb&1)){
      //choose rounding direction in function of the signs of th and tl
			if ((tl>0.0)^(th<0.0))
      	*thpb++;
      else
      	*thpb--;  
    }
  }
  th = *thp;
  free(thp);
  return th;
}

template<class T>
__host__ __device__ T round2odd(const T a, const T b){
  #if defined __CUDA_ARCH__
  	return round2odd_gpu(a, b);
  #else
  	return round2odd_cpu(a, b);
  #endif
}*/

/*template<class T1, class T2, class T3, class TR>
__host__ __device__ TR three_sum(const T1 a, const T2 b, const T3 c){
  TR uh, ul, th, tl;
  uh = two_sum<T2,T3,TR>(b, c, ul);
  th = two_sum<T1,TR,TR>(a, uh, tl);
  tl = round2odd(tl, ul);
 	return (th + tl);
}*/

#endif
