/*
 * multiplication.h
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2014 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */
/*
 * multiplication functions and their specialized definitions
 */

#ifndef _multiplication_h
#define _multiplication_h
#include "addition.h"

/** Multiplies K doubles with L doubles and returns the result on an R double expansion who's terms satisfy |r_i|<=ulp(r_{i-1})
		K - size of x, L - size of y, R - size of r
		Uses an adaption of Priest's multiplication algorithm based on scalar products **/
template <int K, class TK, int L, class TL, int R, class TR>
__host__ __device__ static __forceinline__ void certifiedMul(const TK *x, const TL *y, TR *r){
  TR p[L], e[L], s[2*L], aux[2*K*L];
  int auxS=2*L;
  for(int j=0; j<L; j++)
    p[j] = two_prod<TK,TL,TR>(x[0],y[j],e[j]);
  renorm2L<L,L,TR>(e);
	certifiedAdd<L,TR,L,TR,2*L,TR>(p,e,aux);

  for(int i=1; i<K; i++){
    for(int j=0; j<L; j++)
      p[j] = two_prod<TK,TL,TR>(x[i],y[j],e[j]);
    renorm2L<L,L,TR>(e);
    certifiedAdd<L,TR,L,TR,2*L,TR>(p,e,s);
    auxS = certifiedAdd_4Mul<TR,TR,TR>(s,2*L,aux,auxS,aux);
  }
  if(auxS<R){
    for(int i=0; i<auxS; i++) r[i] = aux[i];
    for(int i=auxS; i<R; i++) r[i] = 0;
  }else
    for(int i=0; i<R; i++) r[i] = aux[i];
}

/** Multiplies K doubles with L doubles and returns the result on an R double expansion
		K - size of x, L - size of y, R - size of z
    Uses a generalization of Bailey's multiplication algorithm that simulates the paper-and-pencil method
		Constraints on the algorithm: K>=L; R<=2*K*L (if R>=2*K*L the extra terms will be 0) **/
template <int K, class TK, int L, class TL, int R, class TR>
__host__ __device__ static __forceinline__ void baileyMul(TK const *x, TL const *y, TR *z){
  TR p, e;
  int i, j, n;
  for(i=K+L-1; i<R-1; i++) z[i] = 0.0;

  //computes the last term of the result z[R-1]
  n = R-1;
  //computes the necessary products (if any) using just simple multiplication
  if (n<L){
    z[n] = FPmul_rn<TK,TL,TR>(x[0], y[n]);
    for(i=1; i<=n; i++) z[n] = FPadd_rn<TR,TR,TR>(z[n], FPmul_rn<TK,TL,TR>(x[i], y[n-i]));
  }else if (n<K){
    z[n] = FPmul_rn<TK,TL,TR>(x[n], y[0]);
    for(j=1; j<L; j++) z[n] = FPadd_rn<TR,TR,TR>(z[n], FPmul_rn<TK,TL,TR>(x[n-j], y[j]));
  }else if (n<K+L-1){
    z[n] = FPmul_rn<TK,TL,TR>(x[n-L+1], y[L-1]);
    for(i=n-L+2; i<K; i++) z[n] = FPadd_rn<TR,TR,TR>(z[n], FPmul_rn<TK,TL,TR>(x[i], y[n-i]));
  }else
    z[n] = 0.0;
  // computes the last R-K elements of the result
  // we will have (K+L-1 - n) products to compute and sum
  for (n=(R>K+L-1)?K+L-2:R-2; n>=K; n--){
    z[n] = two_prod<TK,TL,TR>(x[n-L+1], y[L-1], e);
    for(j=n+1; j<R-1; j++) z[j] = two_sum<TR,TR,TR>(z[j], e, e);
    z[R-1] = FPadd_rn<TR,TR,TR>(z[R-1], e);

    for(i=n-L+2; i<K; i++){
      p = two_prod<TK,TL,TR>(x[i],y[n-i],e);
      for(j=n+1; j<R-1; j++) z[j] = two_sum<TR,TR,TR>(z[j], e, e);
      z[R-1] = FPadd_rn<TR,TR,TR>(z[R-1], e);

      z[n] = two_sum<TR,TR,TR>(z[n], p, e);
      for(j=n+1; j<R-1; j++) z[j] = two_sum<TR,TR,TR>(z[j], e, e);
      z[R-1] = FPadd_rn<TR,TR,TR>(z[R-1], e);
    }
  }
  // computes the elements with the same number of products inside
  // we will have L products to compute and sum
  for (n=(R>K)?K-1:R-2; n>=L; n--){
    z[n] = two_prod<TK,TL,TR>(x[n], y[0], e);
    for(i=n+1; i<R-1; i++) z[i] = two_sum<TR,TR,TR>(z[i], e, e);
    z[R-1] = FPadd_rn<TR,TR,TR>(z[R-1], e);

    for(j=1; j<L; j++){
      p = two_prod<TK,TL,TR>(x[n-j], y[j], e);
      for(i=n+1; i<R-1; i++) z[i] = two_sum<TR,TR,TR>(z[i], e, e);
      z[R-1] = FPadd_rn<TR,TR,TR>(z[R-1], e);

      z[n] = two_sum<TR,TR,TR>(z[n], p, e);
      for(i=n+1; i<R-1; i++) z[i] = two_sum<TR,TR,TR>(z[i], e, e);
      z[R-1] = FPadd_rn<TR,TR,TR>(z[R-1], e);
    }
  }
  // computes the first L doubles of the result
  // we will have (n+1) prducts to compute and sum
  for (n=(R>L)?L-1:R-2; n>=0; n--){
    z[n] = two_prod<TK,TL,TR>(x[0], y[n], e);
    for(j=n+1; j<R-1; j++) z[j] = two_sum<TR,TR,TR>(z[j], e, e);
    z[R-1] = FPadd_rn<TR,TR,TR>(z[R-1], e);

    for(i=1; i<=n; i++){
      p = two_prod<TK,TL,TR>(x[i], y[n-i], e);
      for(j=n+1; j<R-1; j++) z[j] = two_sum<TR,TR,TR>(z[j], e, e);
      z[R-1] = FPadd_rn<TR,TR,TR>(z[R-1], e);

      z[n] = two_sum<TR,TR,TR>(z[n], p, e);
      for(j=n+1; j<R-1; j++) z[j] = two_sum<TR,TR,TR>(z[j], e, e);
      z[R-1] = FPadd_rn<TR,TR,TR>(z[R-1], e);
    }
  }
}

/** Multiplies x and y and returns the result as an expansion in z
		K - size of x, L - size of y, R - size of z
    Uses a generalization of Bailey's multiplication algorithm that simulates the paper-and-pencil method **/
template<int s1, class T1, int s2, class T2, int sR, class TR>
__host__ __device__ static __forceinline__ void truncatedMul(T1 const *x, T2 const *y, TR *z){
  TR aux[sR+1];

  if(s1>=s2) baileyMul<s1,T1,s2,T2,sR+1,TR>(x, y, aux);
  else baileyMul<s2,T2,s1,T1,sR+1,TR>(y, x, aux);
  renorm2L<sR+1,sR,TR>(aux);
  for(int i=0; i<sR; i++) z[i] = aux[i];
}

//*********************************************************//
/** Multiplication algorithm used by the functions based on the Newton-Raphson iteration. The renormalization function does all the subsequent levels **/
template<int s1, class T1, int s2, class T2, int sR, class TR>
__host__ __device__ static __forceinline__ void truncatedMul_4Ntn(T1 const *x, T2 const *y, TR *z){
  TR aux[sR+1];

  if(s1>=s2) baileyMul<s1,T1,s2,T2,sR+1,TR>(x, y, aux);
  else baileyMul<s2,T2,s1,T1,sR+1,TR>(y, x, aux);
  renorm3L<sR+1,sR,TR>(aux);
  for(int i=0; i<sR; i++) z[i] = aux[i];
}

//*********************************************************//
//******** specialized functions for certifiedMul *********//
//*********************************************************//
template <>
__host__ __device__ __forceinline__ void certifiedMul<1,double,1,double,1,double>(const double *x, const double *y, double *r){
	r[0] = FPmul_rn<double,double,double>(x[0], y[0]);
}
template <>
__host__ __device__ __forceinline__ void certifiedMul<1,double,1,double,2,double>(const double *x, const double *y, double *r){
	r[0] = two_prod<double,double,double>(x[0], y[0], r[1]);
}

//*********************************************************//
//********* specialized functions for baileyMul ***********//
//*********************************************************//
template <>
__host__ __device__ __forceinline__ void baileyMul<1,double,1,double,1,double>(double const *x, double const *y, double *z){
	z[0] = FPmul_rn<double,double,double>(x[0], y[0]);
}
template<>
__host__ __device__ __forceinline__ void baileyMul<1,double,1,double,2,double>(double const *x, double const *y, double *z){
  z[0] = two_prod<double,double,double>(x[0], y[0],z[1]);
}
template<>
__host__ __device__ __forceinline__ void baileyMul<2,double,1,double,3,double>(double const *x, double const *y, double *z){
	double e;
  z[1] = two_prod<double,double,double>(x[1], y[0], z[2]);
  z[0] = two_prod<double,double,double>(x[0], y[0], e);
  z[1] = two_sum<double,double,double>(z[1], e, e); z[2] = FPadd_rn<double,double,double>(z[2], e);
}
template<>
__host__ __device__ __forceinline__ void baileyMul<2,double,2,double,3,double>(double const *x, double const *y, double *z){
  double p, e; 
  z[2] = FPmul_rn<double,double,double>(x[1], y[1]);
  z[1] = two_prod<double,double,double>(x[0], y[1], e);
  z[2] = FPadd_rn<double,double,double>(z[2], e);
  p = two_prod<double,double,double>(x[1], y[0], e);
  z[2] = FPadd_rn<double,double,double>(z[2], e);
  z[1] = two_sum<double,double,double>(z[1], p, e); z[2] = FPadd_rn<double,double,double>(z[2], e);
  z[0] = two_prod<double,double,double>(x[0], y[0], e);
  z[1] = two_sum<double,double,double>(z[1], e, e); z[2] = FPadd_rn<double,double,double>(z[2], e);
}
template<>
__host__ __device__ __forceinline__ void baileyMul<3,double,1,double,4,double>(double const *x, double const *y, double *z){
  double e;
  z[2] = two_prod<double,double,double>(x[2], y[0], z[3]);
  z[1] = two_prod<double,double,double>(x[1], y[0], e);
  z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = FPadd_rn<double,double,double>(z[3], e);
  z[0] = two_prod<double,double,double>(x[0], y[0], e);
  z[1] = two_sum<double,double,double>(z[1], e, e); z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = FPadd_rn<double,double,double>(z[3], e);
}
template<>
__host__ __device__ __forceinline__ void baileyMul<3,double,3,double,4,double>(double const *x, double const *y, double *z){
  double p, e;
  z[3] = FPmul_rn<double,double,double>(x[1], y[2]);
  z[3] = FPadd_rn<double,double,double>(z[3], FPmul_rn<double,double,double>(x[2], y[1]));
  z[2] = two_prod<double,double,double>(x[0], y[2], e);
  z[3] = FPadd_rn<double,double,double>(z[3], e);
  p = two_prod<double,double,double>(x[1], y[1], e);
  z[3] = FPadd_rn<double,double,double>(z[3], e);
  z[2] = two_sum<double,double,double>(z[2], p, e); z[3] = FPadd_rn<double,double,double>(z[3], e);
  p = two_prod<double,double,double>(x[2], y[0], e);
  z[3] = FPadd_rn<double,double,double>(z[3], e);
  z[2] = two_sum<double,double,double>(z[2], p, e); z[3] = FPadd_rn<double,double,double>(z[3], e);
  z[1] = two_prod<double,double,double>(x[0], y[1], e);
  z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = FPadd_rn<double,double,double>(z[3], e);
  p = two_prod<double,double,double>(x[1], y[0], e);
  z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = FPadd_rn<double,double,double>(z[3], e);
  z[1] = two_sum<double,double,double>(z[1], p, e); z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = FPadd_rn<double,double,double>(z[3], e);
  z[0] = two_prod<double,double,double>(x[0], y[0], e);
  z[1] = two_sum<double,double,double>(z[1], e, e); z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = FPadd_rn<double,double,double>(z[3], e);
}
template<>
__host__ __device__ __forceinline__ void baileyMul<4,double,1,double,5,double>(double const *x, double const *y, double *z){
  double e;
  z[3] = two_prod<double,double,double>(x[3], y[0], z[4]);
  z[2] = two_prod<double,double,double>(x[2], y[0], e);
  z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  z[1] = two_prod<double,double,double>(x[1], y[0], e);
  z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  z[0] = two_prod<double,double,double>(x[0], y[0], e);
  z[1] = two_sum<double,double,double>(z[1], e, e); z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
}
template<>
__host__ __device__ __forceinline__ void baileyMul<4,double,4,double,5,double>(double const *x, double const *y, double *z){
  double p, e;
  z[4] = FPmul_rn<double,double,double>(x[1], y[3]);
  z[4] = FPadd_rn<double,double,double>(z[4], FPmul_rn<double,double,double>(x[2], y[2]));
  z[4] = FPadd_rn<double,double,double>(z[4], FPmul_rn<double,double,double>(x[3], y[1]));
  z[3] = two_prod<double,double,double>(x[0], y[3], e);
  z[4] = FPadd_rn<double,double,double>(z[4], e);
  p = two_prod<double,double,double>(x[1], y[2], e);
  z[4] = FPadd_rn<double,double,double>(z[4], e);
  z[3] = two_sum<double,double,double>(z[3], p, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  p = two_prod<double,double,double>(x[2], y[1], e);
  z[4] = FPadd_rn<double,double,double>(z[4], e);
  z[3] = two_sum<double,double,double>(z[3], p, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  p = two_prod<double,double,double>(x[3], y[0], e);
  z[4] = FPadd_rn<double,double,double>(z[4], e);
  z[3] = two_sum<double,double,double>(z[3], p, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  z[2] = two_prod<double,double,double>(x[0], y[2], e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  p = two_prod<double,double,double>(x[1], y[1], e);
  z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  z[2] = two_sum<double,double,double>(z[2], p, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  p = two_prod<double,double,double>(x[2], y[0], e);
  z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  z[2] = two_sum<double,double,double>(z[2], p, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  z[1] = two_prod<double,double,double>(x[0], y[1], e);
  z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  p = two_prod<double,double,double>(x[1], y[0], e);
  z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  z[1] = two_sum<double,double,double>(z[1], p, e); z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
  z[0] = two_prod<double,double,double>(x[0], y[0], e);
  z[1] = two_sum<double,double,double>(z[1], e, e); z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = FPadd_rn<double,double,double>(z[4], e);
}
template<>
__host__ __device__ __forceinline__ void baileyMul<8,double,1,double,9,double>(double const *x, double const *y, double *z){
  double e;
  z[7] = FPmul_rn<double,double,double>(x[7], y[0]);
  z[6] = two_prod<double,double,double>(x[6], y[0], e);
  z[7] = FPadd_rn<double,double,double>(z[7], e);
  z[5] = two_prod<double,double,double>(x[5], y[0], e);
  z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = FPadd_rn<double,double,double>(z[7], e);
  z[4] = two_prod<double,double,double>(x[4], y[0], e);
  z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = FPadd_rn<double,double,double>(z[7], e);
  z[3] = two_prod<double,double,double>(x[3], y[0], e);
  z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = FPadd_rn<double,double,double>(z[7], e);
  z[2] = two_prod<double,double,double>(x[2], y[0], e);
  z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e);
  z[7] = FPadd_rn<double,double,double>(z[7], e);
  z[1] = two_prod<double,double,double>(x[1], y[0], e);
  z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e);
  z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = FPadd_rn<double,double,double>(z[7], e);
  z[0] = two_prod<double,double,double>(x[0], y[0], e);
  z[1] = two_sum<double,double,double>(z[1], e, e); z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = two_sum<double,double,double>(z[4], e, e);
  z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = FPadd_rn<double,double,double>(z[7], e);
}
template<>
__host__ __device__ __forceinline__ void baileyMul<8,double,8,double,9,double>(double const *x, double const *y, double *z){
  double p, e;
  z[8] = FPmul_rn<double,double,double>(x[1], y[7]);
  z[8] = FPadd_rn<double,double,double>(z[8], FPmul_rn<double,double,double>(x[2], y[6]));
  z[8] = FPadd_rn<double,double,double>(z[8], FPmul_rn<double,double,double>(x[3], y[5]));
  z[8] = FPadd_rn<double,double,double>(z[8], FPmul_rn<double,double,double>(x[4], y[4]));
  z[8] = FPadd_rn<double,double,double>(z[8], FPmul_rn<double,double,double>(x[5], y[3]));
  z[8] = FPadd_rn<double,double,double>(z[8], FPmul_rn<double,double,double>(x[6], y[2]));
  z[8] = FPadd_rn<double,double,double>(z[8], FPmul_rn<double,double,double>(x[7], y[1]));
  z[7] = two_prod<double,double,double>(x[0], y[7], e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[1], y[6], e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[7] = two_sum<double,double,double>(z[7], p, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[2], y[5], e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[7] = two_sum<double,double,double>(z[7], p, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[3], y[4], e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[7] = two_sum<double,double,double>(z[7], p, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[4], y[3], e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[7] = two_sum<double,double,double>(z[7], p, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[5], y[2], e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[7] = two_sum<double,double,double>(z[7], p, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[6], y[1], e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[7] = two_sum<double,double,double>(z[7], p, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[7], y[0], e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[7] = two_sum<double,double,double>(z[7], p, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[6] = two_prod<double,double,double>(x[0], y[6], e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[1], y[5], e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[6] = two_sum<double,double,double>(z[6], p, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[2], y[4], e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[6] = two_sum<double,double,double>(z[6], p, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[3], y[3], e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[6] = two_sum<double,double,double>(z[6], p, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[4], y[2], e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[6] = two_sum<double,double,double>(z[6], p, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[5], y[1], e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[6] = two_sum<double,double,double>(z[6], p, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[6], y[0], e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[6] = two_sum<double,double,double>(z[6], p, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[5] = two_prod<double,double,double>(x[0], y[5], e);
  z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[1], y[4], e);
  z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[5] = two_sum<double,double,double>(z[5], p, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[2], y[3], e);
  z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[5] = two_sum<double,double,double>(z[5], p, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[3], y[2], e);
  z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[5] = two_sum<double,double,double>(z[5], p, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[4], y[1], e);
  z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[5] = two_sum<double,double,double>(z[5], p, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[5], y[0], e);
  z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[5] = two_sum<double,double,double>(z[5], p, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[4] = two_prod<double,double,double>(x[0], y[4], e);
  z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[1], y[3], e);
  z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[4] = two_sum<double,double,double>(z[4], p, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[2], y[2], e);
  z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[4] = two_sum<double,double,double>(z[4], p, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[3], y[1], e);
  z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[4] = two_sum<double,double,double>(z[4], p, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[4], y[0], e);
  z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[4] = two_sum<double,double,double>(z[4], p, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[3] = two_prod<double,double,double>(x[0], y[3], e);
  z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[1], y[2], e);
  z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[3] = two_sum<double,double,double>(z[3], p, e); z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[2], y[1], e);
  z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[3] = two_sum<double,double,double>(z[3], p, e); z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[3], y[0], e);
  z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e);
  z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[3] = two_sum<double,double,double>(z[3], p, e); z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[2] = two_prod<double,double,double>(x[0], y[2], e);
  z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[1], y[1], e);
  z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[2] = two_sum<double,double,double>(z[2], p, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e);
  z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[2], y[0], e);
  z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e);
  z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[2] = two_sum<double,double,double>(z[2], p, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e);
  z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[1] = two_prod<double,double,double>(x[0], y[1], e);
  z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e);
  z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  p = two_prod<double,double,double>(x[1], y[0], e);
  z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = two_sum<double,double,double>(z[4], e, e); z[5] = two_sum<double,double,double>(z[5], e, e);
  z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[1] = two_sum<double,double,double>(z[1], p, e); z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = two_sum<double,double,double>(z[4], e, e);
  z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
  z[0] = two_prod<double,double,double>(x[0], y[0], e);
  z[1] = two_sum<double,double,double>(z[1], e, e); z[2] = two_sum<double,double,double>(z[2], e, e); z[3] = two_sum<double,double,double>(z[3], e, e); z[4] = two_sum<double,double,double>(z[4], e, e);
  z[5] = two_sum<double,double,double>(z[5], e, e); z[6] = two_sum<double,double,double>(z[6], e, e); z[7] = two_sum<double,double,double>(z[7], e, e); z[8] = FPadd_rn<double,double,double>(z[8], e);
}

//*********************************************************//
//******** specialized functions for truncatedMul *********//
//*********************************************************//
template<>
__host__ __device__ __forceinline__ void truncatedMul<1,double,1,double,1,double>(double const *x, double const *y, double *z){
	z[0] = FPmul_rn<double,double,double>(x[0], y[0]);
}
template<>
__host__ __device__ __forceinline__ void truncatedMul<1,double,1,double,2,double>(double const *x, double const *y, double *z){
 z[0] = two_prod<double,double,double>(x[0], y[0],z[1]);
}

//*********************************************************//
//***** specialized functions for truncatedMul_4Ntn *******//
//*********************************************************//
template<>
__host__ __device__ __forceinline__ void truncatedMul_4Ntn<1,double,1,double,1,double>(double const *x, double const *y, double *z){
  z[0] = FPmul_rn<double,double,double>(x[0], y[0]);
}
template<>
__host__ __device__ __forceinline__ void truncatedMul_4Ntn<1,double,1,double,2,double>(double const *x, double const *y, double *z){
 z[0] = two_prod<double,double,double>(x[0], y[0],z[1]);
}

#endif
