/*
 * newton.h
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2014 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */
/*
 * newton iteration based functions and their specialized definitions
 */

#ifndef _newton_h
#define _newton_h
#include "multiplication.h"

/***************************************************************
 ----------------- DIVISION ALGORITHMS -------------------
 ***************************************************************/
/** Computes the reciprocal of the array x of size sX and puts the result in the array res of size sR
    Uses a recursive Newton iteration method for 1/y, with initial guess q_0;
    The iteration is q_{n+1} = q_n(2-yq_n) **/
/* downward version --> not used in the CAMPARY library */
template <int sX, class T>
__host__ __device__ static void __forceinline__ invNewtonDegenerate(T const *x, T *res){
	// computes an approximation of the reciprocal, 1/x_0
  res[0] = 1/x[0];
}
template<int sX, class T1, int sR, class TR>
__host__ __device__ static void __forceinline__ invNewtonDown(T1 const *x, TR *res){
  if(sR < 1) return;
  else if(sR == 1){
    invNewtonDegenerate<sX,TR>(x, res);
    return;
  }else{
    const int i = (sR-1)/2 +1;
    if(i==1) invNewtonDegenerate<sX,TR>(x, res); 
    else invNewtonDown<sX,T1,i,TR>(x, res);

    TR aux[sR];
    const int ii = 2*i;
    truncatedMul_4Ntn<i,TR, (ii>sX)?sX:ii,T1, ii,TR>(res, x, aux);
    for(int j=0; j<sR; j++) aux[j] = -aux[j];

		const TR two[] = {2.}; 
    truncatedAdd_4Ntn<ii,TR, 1,TR, ii,TR>(aux, two, aux);
    truncatedMul_4Ntn<ii,TR, i,TR, ii,TR>(aux, res, res);
  }
}

/** Computes the reciprocal of the array x of size sX and puts the result in the array res of size sR
    Uses a recursive Newton iteration method for 1/y, with initial guess q_0;
    The iteration is q_{n+1} = q_n(2-yq_n) 
		Constraints: we assume that sR and sX >= 1 **/
/* upward version --> used in the CAMPARY library */
template<int sX, class T1, int sR, class TR, int pDone>
__host__ __device__ static void __forceinline__ invNewtonUp_aux(T1 const *x, TR *res){
	// pDone is the precision we have already computed in res
  const int p  = (2*pDone>sR) ? sR : 2*pDone;
  const TR two[] = {2.};
  TR aux[p];

  truncatedMul_4Ntn<pDone, TR, (p>sX)?sX:p, T1, p, TR>(res, x, aux);
  for(int j=0; j<p; j++) aux[j] = -aux[j];

  truncatedAdd_4Ntn<p, TR, 1, TR, p, TR>(aux, two, aux);
  truncatedMul_4Ntn<p, TR, pDone, TR, p, TR>(aux, res, res);

  if(p<sR) invNewtonUp_aux<sX,T1,sR,TR,p>(x, res);
}
template<int sX, class T1, int sR, class TR>
__host__ __device__ static void __forceinline__ invNewtonUp(T1 const *x, TR *res){
	// starts by computing an approximation of the reciprocal, 1/x_0
  res[0] = 1./x[0];
  if(sR > 1) invNewtonUp_aux<sX,T1,sR,TR,1>(x, res);
}

/** Computes the division of the array x by the array y and puts the result in the array res
		Firstly computes the reciprocal of y and after that multiplies the result by x **/
template<int sX, class TX, int sizeY, class TY, int sR, class TR>
__host__ __device__ static void __forceinline__ divNewton(TX const *x, TY const *y, TR *res){
  TR aux[sR];
  invNewtonUp<sizeY,TY,sR,TR>(y, aux);
  if(sX > sR) truncatedMul<sX,TX,sR,TR,sR,TR>(x, aux, res);
  else truncatedMul<sR,TR,sX,TX,sR,TR>(aux, x, res);
}


/***************************************************************
 ----------------- SQUARE ROOT ALGORITHMS -------------------
 ***************************************************************/
/** Computes the reciprocal of the square root of array x of size sX and puts the result in the array res of size sR
    Uses a recursive Newton iteration method for 1/sqrt(y), with initial guess q_0;
    The 'division-free' iteration is q_{n+1} = q_n(3-x*q_n^2)/2
		Constraints: we assume that sR and sX >= 1 **/
template<int sX, class T1, int sR, class TR, int pDone>
__host__ __device__ static void __forceinline__ sqrtInvNewton_aux(T1 const *x, TR *res){
  const int p  = (2*pDone>sR) ? sR : 2*pDone;
  const TR three[] = {3.};
  TR aux[p];

  truncatedMul_4Ntn<(p>sX)?sX:p, T1, pDone, TR, p, TR>(x, res, aux);
  truncatedMul_4Ntn<p, TR, pDone, TR, p, TR>(aux, res, aux);

  for(int j=0; j<p; j++) aux[j] = -aux[j];
  truncatedAdd_4Ntn<p, TR, 1, TR, p, TR>(aux, three, aux);
  truncatedMul_4Ntn<p, TR, pDone, TR, p, TR>(aux, res, res);
  for(int j=0; j<p; j++) res[j] *= 0.5;

  if(p<sR) sqrtInvNewton_aux<sX,T1,sR,TR,p>(x, res);
}
template<int sX, class T1, int sR, class TR>
__host__ __device__ static void __forceinline__ sqrtInvNewton(T1 const *x, TR *res){
  if(x[0]<0.) for(int i=0; i<sR; i++) res[i] = nan("");
  else if(x[0]==0.) for(int i=0; i<sR; i++) res[i] = 0.;
  else{
    res[0] = 1./sqrt(x[0]);
    if(sR>1) sqrtInvNewton_aux<sX,T1,sR,TR,1>(x, res);
  }
}

/** Computes the square root of array x of size sX and puts the result in the array res of size sR
		Firstly computes the reciprocal of the square root of x and after that multiplies the result by x **/
template<int sX, class T1, int sR, class TR>
__host__ __device__ static void __forceinline__ sqrtDirectNewton(T1 const *x, TR *res){
  TR aux[sR];
  sqrtInvNewton<sX, T1, sR, TR>(x, aux);
  if(sX > sR) truncatedMul<sX,T1,sR,TR,sR,TR>(x, aux, res);
  else truncatedMul<sR,TR,sX,T1,sR,TR>(aux, x, res);
}

/** Computes the square root of array x of size sX and puts the result in the array res of size sR
    Uses a recursive Newton iteration method for sqrt(y), with initial guess q_0;
    The 'Heron' iteration is q_{n+1} = (q_n+x/q_n)/2
    Constraints: we assume that sR and sX >= 1 **/
template<int sX, class T1, int sR, class TR, int pDone>
__host__ __device__ static void __forceinline__ sqrtHeron_aux(T1 const *x, TR *res){
  const int p  = (2*pDone>sR) ? sR : 2*pDone;
  TR aux[p];

  divNewton<(p>sX)?sX:p, T1, pDone, TR, p, TR>(x, res, aux);
  truncatedAdd_4Ntn<pDone, TR, p, TR, p, TR>(res, aux, res);

  for(int j=0; j<p; j++) res[j] *= 0.5;

  if(p<sR) sqrtHeron_aux<sX,TR,sR,TR,p>(x, res);
}
template<int sX, class T1, int sR, class TR>
__host__ __device__ static void __forceinline__ sqrtHeron(T1 const *x, TR *res){
  if(x[0]<0.) for(int i=0; i<sR; i++) res[i] = nan("");
  else if(x[0]==0.) for(int i=0; i<sR; i++) res[i] = 0.;
  else{
    res[0] = sqrt(x[0]);
    if(sR>1) sqrtHeron_aux<sX,T1,sR,TR,1>(x, res);
  }
}

//*********************************************************//
//****** specialized functions for Newton functions *******//
//*********************************************************//
template<>
__host__ __device__ void __forceinline__ invNewtonUp<1,double,1,double>(double const *x, double *res){ res[0] = 1./x[0]; }

template<>
__host__ __device__ void __forceinline__ divNewton<1,double,1,double,1,double>(double const *x, double const *y, double *res){ res[0] = x[0] / y[0]; }

template<>
__host__ __device__ void __forceinline__ sqrtInvNewton<1,double,1,double>(double const *x, double *res){ res[0] = 1/sqrt(x[0]); }

template<>
__host__ __device__ void __forceinline__ sqrtDirectNewton<1,double,1,double>(double const *x, double *res){ res[0] = sqrt(x[0]); }

template<>
__host__ __device__ void __forceinline__ sqrtHeron<1,double,1,double>(double const *x, double *res){ res[0] = sqrt(x[0]); }

#endif
