/*
 * renorm.h
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2014 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */
/*
 * RENORMALISATION ALGORITHMS and their specialized definitions
 */

#ifndef _renorm_h
#define _renorm_h
#include "errorFreeTransf.h"
#include <stdio.h>
#include <math.h>
#include <fenv.h>
#include <cstdio>
#include <iostream>
using namespace std;

/** Algorithm for normalizing the array x. The result satisfies |x_i|<=ulp(x_{i-1}).
		sX is the size of array x and sR is the required size for the normalized expansion
		Requirement: sR <= sX **/
template<int sX, int sR, class T>
__host__ __device__ static __forceinline__ void renorm2L(T *x){
  T pr, f[sX];

  // first level - VecSum RL
  pr = quick_two_sum<T,T,T>(x[sX-2], x[sX-1], f[sX-1]);
  for(int i=sX-3; i>0; i--) pr = quick_two_sum<T,T,T>(x[i], pr, f[i+1]);
  x[0] = quick_two_sum<T,T,T>(x[0], pr, pr);

  // second level - VecSumErr_branch LR
  int ind=0, i=2;
  if(pr==0.0) pr = x[0]; else ind++;
	while(ind<sR && i<sX){
		x[ind] = quick_two_sum<T,T,T>(pr, f[i], pr); i++;
    if(pr==0.0) pr = x[ind]; else ind++;
  }
	if(i==sX && ind<sR && pr!=0.0){ x[ind] = pr; ind++; }	
	for(i=ind; i<sR; i++) x[i] = 0.;
	for(i=sR; i<sX; i++) x[i] = 0.;
}

/** Algorithm for normalizing the array x. The result is a P-nonoverlapping expansion of size sR.
		sX is the size of array x and sR is the required size for the normalized expansion
    Requirement: sR <= sX **/
template<int sX, int sR, class T>
__host__ __device__ static __forceinline__ void renorm3L(T *x){
  T pr, f[(sR<sX)?sX:sR];

  // first level - VecSum RL
  pr = quick_two_sum<T,T,T>(x[sX-2], x[sX-1], f[sX-1]);
  for(int i=sX-3; i>0; i--) pr = quick_two_sum<T,T,T>(x[i], pr, f[i+1]);
  f[0] = quick_two_sum<T,T,T>(x[0], pr, pr);

  // second level - VecSumErr_branch LR
  int ind=0, i=2;
  if(pr==0.0) pr = f[0]; else ind++;
  while(ind<((sR<sX)?sR+1:sR) && i<sX){
    f[ind] = quick_two_sum<T,T,T>(pr, f[i], pr); i++;
    if(pr==0.0) pr = f[ind]; else ind++;
  }
  if(i==sX && ind<((sR<sX)?sR+1:sR) && pr!=0.0){ f[ind] = pr; ind++; }
  for(i=ind; i<((sR<sX)?sR+1:sR); i++) f[i] = 0.;

  // third level and further - VecSumErr LR
  for(int j=0; j<((sR<sX)?sR-1:sX-2); j++){
    x[j] = quick_two_sum<T,T,T>(f[j], f[j+1], pr);
    for(int i=j+1; i<((sR<sX)?sR-1:sX-2); i++) f[i] = quick_two_sum<T,T,T>(pr, f[i+1], pr);
    f[(sR<sX)?sR-1:sX-2] = quick_two_sum<T,T,T>(pr, f[(sR<sX)?sR:sX-1], f[(sR<sX)?sR:sX-1]);
  }
  x[(sR<sX)?sR-1:sX-2] = f[(sR<sX)?sR-1:sX-2];
  x[(sR<sX)?sR-1:sX-1] = f[(sR<sX)?sR-1:sX-1];

  for(i=sR; i<sX; i++) x[i] = 0.;
}

//*********************************************************//
/** Renormalization algorithm used by the certifiedMul function
    Returns the normalized result in the array r and returns the new size of r **/
template<class T>
__host__ __device__ static __forceinline__ int renorm2L_4Mul(const T *x, const int sX, T *r){
  T pr, *f = new T[sX];

  if(sX <= 0) return 0;
  else if(sX == 1){ r[0] = x[0]; return 1;}
  else if(sX == 2){
    r[0] = quick_two_sum<T,T,T>(x[0], x[1], r[1]);
    if(r[1] == 0) return 1;
    else return 2;
  }

  // first level - VecSum RL
  pr = quick_two_sum<T,T,T>(x[sX-2], x[sX-1], f[sX-1]);
  for(int i=sX-3; i>0; i--) pr = quick_two_sum<T,T,T>(x[i], pr, f[i+1]);
  r[0] = quick_two_sum<T,T,T>(x[0], pr, pr);

  // second level - VecSumErr_branch LR
  int newS=0;
  if(pr==0.0) pr = r[0]; else newS++;
  for (int i=2; i<sX; i++){
    r[newS] = quick_two_sum<T,T,T>(pr, f[i], pr);
    if(pr==0.0) pr = r[newS]; else newS++;
  }
  if(pr!=0.0){ r[newS] = pr; newS++; }
  return newS;
}

//*********************************************************//
//********** specialized functions for renorm2L ***********//
//*********************************************************//
template<>
__host__ __device__ __forceinline__ void renorm2L<1,1,double>(double *x){ }
template<>
__host__ __device__ __forceinline__ void renorm2L<2,2,double>(double *x){
	x[0] = quick_two_sum<double,double,double>(x[0], x[1], x[1]);
}
template<>
__host__ __device__ __forceinline__ void renorm2L<2,1,double>(double *x){
  x[0] = FPadd_rn<double,double,double>(x[0], x[1]); x[1] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm2L<3,2,double>(double *x){
  double pr, f[3];
  pr = quick_two_sum<double,double,double>(x[1],x[2],f[2]);
  x[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = x[0]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[2],x[ind+1]);
	x[2] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm2L<4,2,double>(double *x){
  double pr, f[4];
  pr = quick_two_sum<double,double,double>(x[2],x[3],f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  x[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = x[0]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[3],x[ind+1]);
  x[2] = 0.; x[3] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm2L<4,3,double>(double *x){
  double pr, f[4];
  pr = quick_two_sum<double,double,double>(x[2],x[3],f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  x[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = x[0]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[3],x[ind+1]);
  for(int i=ind+2; i<3; i++) x[i] = 0.;
	x[3] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm2L<5,4,double>(double *x){
	double pr, f[5];
  pr = quick_two_sum<double,double,double>(x[3],x[4],f[4]);
  pr = quick_two_sum<double,double,double>(x[2],pr,f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  x[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = x[0]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[3],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[4],x[ind+1]);
  for(int i=ind+2; i<4; i++) x[i] = 0.;
	x[4] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm2L<6,3,double>(double *x){
  double pr, f[6];
  pr = quick_two_sum<double,double,double>(x[4],x[5],f[5]);
  pr = quick_two_sum<double,double,double>(x[3],pr,f[4]);
  pr = quick_two_sum<double,double,double>(x[2],pr,f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  x[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = x[0]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[3],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[4],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[5],x[ind+1]);
  for(int i=ind+2; i<3; i++) x[i] = 0.;
  x[3] = 0.; x[4] = 0.; x[5] = 0.; 
}
template<>
__host__ __device__ __forceinline__ void renorm2L<8,4,double>(double *x){
  double pr, f[8];
  pr = quick_two_sum<double,double,double>(x[6],x[7],f[7]);
  pr = quick_two_sum<double,double,double>(x[5],pr,f[6]);
  pr = quick_two_sum<double,double,double>(x[4],pr,f[5]);
  pr = quick_two_sum<double,double,double>(x[3],pr,f[4]);
  pr = quick_two_sum<double,double,double>(x[2],pr,f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  x[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = x[0]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[3],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[4],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[5],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[6],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[7],x[ind+1]);
  for(int i=ind+2; i<4; i++) x[i] = 0.;
  x[4] = 0.; x[5] = 0.; x[6] = 0.; x[7] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm2L<9,8,double>(double *x){
	double pr, f[9];
  pr = quick_two_sum<double,double,double>(x[7],x[8],f[8]);
  pr = quick_two_sum<double,double,double>(x[6],pr,f[7]);
  pr = quick_two_sum<double,double,double>(x[5],pr,f[6]);
  pr = quick_two_sum<double,double,double>(x[4],pr,f[5]);
  pr = quick_two_sum<double,double,double>(x[3],pr,f[4]);
  pr = quick_two_sum<double,double,double>(x[2],pr,f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  x[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = x[0]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[3],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[4],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[5],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[6],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[7],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[8],x[ind+1]);
  for(int i=ind+2; i<8; i++) x[i] = 0.;
  x[8] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm2L<16,8,double>(double *x){
	double pr, f[16];
	pr = quick_two_sum<double,double,double>(x[14],x[15],f[15]);
  pr = quick_two_sum<double,double,double>(x[13],pr,f[14]);
  pr = quick_two_sum<double,double,double>(x[12],pr,f[13]);
  pr = quick_two_sum<double,double,double>(x[11],pr,f[12]);
  pr = quick_two_sum<double,double,double>(x[10],pr,f[11]);
  pr = quick_two_sum<double,double,double>(x[9],pr,f[10]);
  pr = quick_two_sum<double,double,double>(x[8],pr,f[9]);
  pr = quick_two_sum<double,double,double>(x[7],pr,f[8]);
  pr = quick_two_sum<double,double,double>(x[6],pr,f[7]);
  pr = quick_two_sum<double,double,double>(x[5],pr,f[6]);
  pr = quick_two_sum<double,double,double>(x[4],pr,f[5]);
  pr = quick_two_sum<double,double,double>(x[3],pr,f[4]);
  pr = quick_two_sum<double,double,double>(x[2],pr,f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  x[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = x[0]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[3],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[4],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[5],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[6],pr);
  if(pr==0) pr = x[ind]; else ind++;
	x[ind] = quick_two_sum<double,double,double>(pr,f[7],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[8],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[9],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[10],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[11],pr);
  if(pr==0) pr = x[ind]; else ind++;
	x[ind] = quick_two_sum<double,double,double>(pr,f[12],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[13],pr);
  if(pr==0) pr = x[ind]; else ind++;
	x[ind] = quick_two_sum<double,double,double>(pr,f[14],pr);
  if(pr==0) pr = x[ind]; else ind++;
  x[ind] = quick_two_sum<double,double,double>(pr,f[15],x[ind+1]);
  for(int i=ind+2; i<8; i++) x[i] = 0.;
  x[8] = 0.; x[9] = 0.; x[10] = 0.; x[11] = 0.;
	x[12] = 0.; x[13] = 0.; x[14] = 0.; x[15] = 0.;
}

//*********************************************************//
//********** specialized functions for renorm3L ***********//
//*********************************************************//
template<>
__host__ __device__ __forceinline__ void renorm3L<1,1,double>(double *x){ }
template<>
__host__ __device__ __forceinline__ void renorm3L<2,2,double>(double *x){ x[0] = quick_two_sum<double,double,double>(x[0], x[1], x[1]); }
template<>
__host__ __device__ __forceinline__ void renorm3L<2,1,double>(double *x){ x[0] = FPadd_rn<double,double,double>(x[0], x[1]); x[1] = 0.; }
template<>
__host__ __device__ __forceinline__ void renorm3L<3,2,double>(double *x){
  double pr, f[3]; 
  pr = quick_two_sum<double,double,double>(x[1],x[2],f[2]);
  f[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = f[0]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[2],f[ind+1]);
  for(int i=ind+2; i<3; i++) f[i] = 0.0;

  x[0] = quick_two_sum<double,double,double>(f[0],f[1],pr);
  x[1] = FPadd_rn<double,double,double>(pr,f[2]);
	x[2] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm3L<4,2,double>(double *x){
  double pr, f[4];
  pr = quick_two_sum<double,double,double>(x[3],x[2],f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  f[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = f[0]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[3],f[ind+1]);
  for(int i=ind+2; i<3; i++) f[i] = 0.0;

  x[0] = quick_two_sum<double,double,double>(f[0],f[1],pr);
	x[1] = FPadd_rn<double,double,double>(pr,f[2]);
	x[2] = 0.; x[3] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm3L<4,3,double>(double *x){
  double pr, f[4];
  pr = quick_two_sum<double,double,double>(x[3],x[2],f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  f[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = f[0]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[3],f[ind+1]);
  for(int i=ind+2; i<4; i++) f[i] = 0.0;

  x[0] = quick_two_sum<double,double,double>(f[0],f[1],pr);
  f[1] = quick_two_sum<double,double,double>(pr,f[2],pr);
  f[2] = quick_two_sum<double,double,double>(pr,f[3],f[3]);

  x[1] = quick_two_sum<double,double,double>(f[1],f[2],pr);
  x[2] = FPadd_rn<double,double,double>(f[3],pr);
	x[3] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm3L<5,4,double>(double *x){
  double pr, f[5];
  pr = quick_two_sum<double,double,double>(x[3],x[4],f[4]);
  pr = quick_two_sum<double,double,double>(x[2],pr,f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  f[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = f[0]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[3],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[4],f[ind+1]);
  for(int i=ind+2; i<5; i++) f[i] = 0.0;

  x[0] = quick_two_sum<double,double,double>(f[0],f[1],pr);
  f[1] = quick_two_sum<double,double,double>(pr,f[2],pr);
  f[2] = quick_two_sum<double,double,double>(pr,f[3],pr);
  f[3] = quick_two_sum<double,double,double>(pr,f[4],f[4]);

  x[1] = quick_two_sum<double,double,double>(f[1],f[2],pr);
  f[2] = quick_two_sum<double,double,double>(pr,f[3],pr);
  f[3] = quick_two_sum<double,double,double>(pr,f[4],f[4]);

  x[2] = quick_two_sum<double,double,double>(f[2],f[3],pr);
  x[3] = FPadd_rn<double,double,double>(f[4],pr);
	x[4] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm3L<6,3,double>(double *x){
  double pr, f[6];
  pr = quick_two_sum<double,double,double>(x[4],x[5],f[5]);
	pr = quick_two_sum<double,double,double>(x[3],pr,f[4]);
  pr = quick_two_sum<double,double,double>(x[2],pr,f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  f[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = f[0]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[3],pr);
	if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[4],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[5],f[ind+1]);
  for(int i=ind+2; i<4; i++) f[i] = 0.0;

  x[0] = quick_two_sum<double,double,double>(f[0],f[1],pr);
  f[1] = quick_two_sum<double,double,double>(pr,f[2],pr);
  f[2] = quick_two_sum<double,double,double>(pr,f[3],f[3]);

  x[1] = quick_two_sum<double,double,double>(f[1],f[2],pr);
  x[2] = FPadd_rn<double,double,double>(pr,f[3]);
  x[3] = 0.; x[4] = 0.; x[5] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm3L<8,4,double>(double *x){
  double pr, f[8];
  pr = quick_two_sum<double,double,double>(x[6],x[7],f[7]);
	pr = quick_two_sum<double,double,double>(x[5],pr,f[6]);
  pr = quick_two_sum<double,double,double>(x[4],pr,f[5]);
  pr = quick_two_sum<double,double,double>(x[3],pr,f[4]);
  pr = quick_two_sum<double,double,double>(x[2],pr,f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  f[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = f[0]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[3],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[4],pr);
	if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[5],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[6],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[7],f[ind+1]);
  for(int i=ind+2; i<5; i++) f[i] = 0.0;

	x[0] = quick_two_sum<double,double,double>(f[0],f[1],pr);
  f[1] = quick_two_sum<double,double,double>(pr,f[2],pr);
  f[2] = quick_two_sum<double,double,double>(pr,f[3],pr);
  f[3] = quick_two_sum<double,double,double>(pr,f[4],f[4]);

  x[1] = quick_two_sum<double,double,double>(f[1],f[2],pr);
  f[2] = quick_two_sum<double,double,double>(pr,f[3],pr);
  f[3] = quick_two_sum<double,double,double>(pr,f[4],f[4]);

  x[2] = quick_two_sum<double,double,double>(f[2],f[3],pr);
  x[3] = FPadd_rn<double,double,double>(f[4],pr);
  x[4] = 0.; x[5] = 0.; x[6] = 0.; x[7] = 0.;
}
template<>
__host__ __device__ __forceinline__ void renorm3L<9,8,double>(double *x){
  double pr, f[9];
  pr = quick_two_sum<double,double,double>(x[7],x[8],f[8]);
  pr = quick_two_sum<double,double,double>(x[6],pr,f[7]);
  pr = quick_two_sum<double,double,double>(x[5],pr,f[6]);
  pr = quick_two_sum<double,double,double>(x[4],pr,f[5]);
  pr = quick_two_sum<double,double,double>(x[3],pr,f[4]);
  pr = quick_two_sum<double,double,double>(x[2],pr,f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  f[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = f[0]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[3],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[4],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[5],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[6],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[7],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[8],f[ind+1]);
  for(int i=ind+2; i<9; i++) f[i] = 0.0;

  x[0] = quick_two_sum<double,double,double>(f[0],f[1],pr);
  f[1] = quick_two_sum<double,double,double>(pr,f[2],pr);
  f[2] = quick_two_sum<double,double,double>(pr,f[3],pr);
  f[3] = quick_two_sum<double,double,double>(pr,f[4],pr);
  f[4] = quick_two_sum<double,double,double>(pr,f[5],pr);
  f[5] = quick_two_sum<double,double,double>(pr,f[6],pr);
  f[6] = quick_two_sum<double,double,double>(pr,f[7],pr);
  f[7] = quick_two_sum<double,double,double>(pr,f[8],f[8]);

  x[1] = quick_two_sum<double,double,double>(f[1],f[2],pr);
  f[2] = quick_two_sum<double,double,double>(pr,f[3],pr);
  f[3] = quick_two_sum<double,double,double>(pr,f[4],pr);
  f[4] = quick_two_sum<double,double,double>(pr,f[5],pr);
  f[5] = quick_two_sum<double,double,double>(pr,f[6],pr);
  f[6] = quick_two_sum<double,double,double>(pr,f[7],pr);
  f[7] = quick_two_sum<double,double,double>(pr,f[8],f[8]);

  x[2] = quick_two_sum<double,double,double>(f[2],f[3],pr);
  f[3] = quick_two_sum<double,double,double>(pr,f[4],pr);
  f[4] = quick_two_sum<double,double,double>(pr,f[5],pr);
  f[5] = quick_two_sum<double,double,double>(pr,f[6],pr);
  f[6] = quick_two_sum<double,double,double>(pr,f[7],pr);
  f[7] = quick_two_sum<double,double,double>(pr,f[8],f[8]);

  x[3] = quick_two_sum<double,double,double>(f[3],f[4],pr);
  f[4] = quick_two_sum<double,double,double>(pr,f[5],pr);
  f[5] = quick_two_sum<double,double,double>(pr,f[6],pr);
  f[6] = quick_two_sum<double,double,double>(pr,f[7],pr);
  f[7] = quick_two_sum<double,double,double>(pr,f[8],f[8]);

  x[4] = quick_two_sum<double,double,double>(f[4],f[5],pr);
  f[5] = quick_two_sum<double,double,double>(pr,f[6],pr);
  f[6] = quick_two_sum<double,double,double>(pr,f[7],pr);
  f[7] = quick_two_sum<double,double,double>(pr,f[8],f[8]);

  x[5] = quick_two_sum<double,double,double>(f[5],f[6],pr);
  f[6] = quick_two_sum<double,double,double>(pr,f[7],pr);
  f[7] = quick_two_sum<double,double,double>(pr,f[8],f[8]);

  x[6] = quick_two_sum<double,double,double>(f[6],f[7],pr);
  x[7] = FPadd_rn<double,double,double>(f[8],pr);
	x[8] = 0.;
}

template<>
__host__ __device__ __forceinline__ void renorm3L<16,8,double>(double *x){
  double pr, f[17];
  pr = quick_two_sum<double,double,double>(x[14],x[15],f[15]);
  pr = quick_two_sum<double,double,double>(x[13],pr,f[14]);
  pr = quick_two_sum<double,double,double>(x[12],pr,f[13]);
  pr = quick_two_sum<double,double,double>(x[11],pr,f[12]);
  pr = quick_two_sum<double,double,double>(x[10],pr,f[11]);
  pr = quick_two_sum<double,double,double>(x[9],pr,f[10]);
  pr = quick_two_sum<double,double,double>(x[8],pr,f[9]);
  pr = quick_two_sum<double,double,double>(x[7],pr,f[8]);
  pr = quick_two_sum<double,double,double>(x[6],pr,f[7]);
  pr = quick_two_sum<double,double,double>(x[5],pr,f[6]);
  pr = quick_two_sum<double,double,double>(x[4],pr,f[5]);
  pr = quick_two_sum<double,double,double>(x[3],pr,f[4]);
  pr = quick_two_sum<double,double,double>(x[2],pr,f[3]);
  pr = quick_two_sum<double,double,double>(x[1],pr,f[2]);
  f[0] = quick_two_sum<double,double,double>(x[0],pr,pr);

  int ind=0;
  if(pr==0) pr = f[0]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[2],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[3],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[4],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[5],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[6],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[7],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[8],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[9],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[10],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[11],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[12],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[13],pr);
  if(pr==0) pr = f[ind]; else ind++;
  f[ind] = quick_two_sum<double,double,double>(pr,f[14],pr);
  if(pr==0) pr = f[ind]; else{ind++;}
  f[ind] = quick_two_sum<double,double,double>(pr,f[15],f[ind+1]);
  for(int i=ind+2; i<9; i++) f[i] = 0.0;
	
  x[0] = quick_two_sum<double,double,double>(f[0],f[1],pr);
  f[1] = quick_two_sum<double,double,double>(pr,f[2],pr);
  f[2] = quick_two_sum<double,double,double>(pr,f[3],pr);
  f[3] = quick_two_sum<double,double,double>(pr,f[4],pr);
  f[4] = quick_two_sum<double,double,double>(pr,f[5],pr);
  f[5] = quick_two_sum<double,double,double>(pr,f[6],pr);
  f[6] = quick_two_sum<double,double,double>(pr,f[7],pr);
  f[7] = quick_two_sum<double,double,double>(pr,f[8],f[8]);

  x[1] = quick_two_sum<double,double,double>(f[1],f[2],pr);
  f[2] = quick_two_sum<double,double,double>(pr,f[3],pr);
  f[3] = quick_two_sum<double,double,double>(pr,f[4],pr);
  f[4] = quick_two_sum<double,double,double>(pr,f[5],pr);
  f[5] = quick_two_sum<double,double,double>(pr,f[6],pr);
  f[6] = quick_two_sum<double,double,double>(pr,f[7],pr);
  f[7] = quick_two_sum<double,double,double>(pr,f[8],f[8]);

  x[2] = quick_two_sum<double,double,double>(f[2],f[3],pr);
  f[3] = quick_two_sum<double,double,double>(pr,f[4],pr);
  f[4] = quick_two_sum<double,double,double>(pr,f[5],pr);
  f[5] = quick_two_sum<double,double,double>(pr,f[6],pr);
  f[6] = quick_two_sum<double,double,double>(pr,f[7],pr);
  f[7] = quick_two_sum<double,double,double>(pr,f[8],f[8]);

  x[3] = quick_two_sum<double,double,double>(f[3],f[4],pr);
  f[4] = quick_two_sum<double,double,double>(pr,f[5],pr);
  f[5] = quick_two_sum<double,double,double>(pr,f[6],pr);
  f[6] = quick_two_sum<double,double,double>(pr,f[7],pr);
  f[7] = quick_two_sum<double,double,double>(pr,f[8],f[8]);

  x[4] = quick_two_sum<double,double,double>(f[4],f[5],pr);
  f[5] = quick_two_sum<double,double,double>(pr,f[6],pr);
  f[6] = quick_two_sum<double,double,double>(pr,f[7],pr);
  f[7] = quick_two_sum<double,double,double>(pr,f[8],f[8]);

  x[5] = quick_two_sum<double,double,double>(f[5],f[6],pr);
  f[6] = quick_two_sum<double,double,double>(pr,f[7],pr);
  f[7] = quick_two_sum<double,double,double>(pr,f[8],f[8]);

  x[6] = quick_two_sum<double,double,double>(f[6],f[7],pr);
  x[7] = FPadd_rn<double,double,double>(f[8],pr);
  x[8] = 0.; x[9] = 0.; x[10] = 0.; x[11] = 0.;
  x[12] = 0.; x[13] = 0.; x[14] = 0.; x[15] = 0.;
}

#endif
